/****************************************************************/
/* avltree.c - AVL Binary Search Tree Source			*/
/****************************************************************/
/*
 *  (c)2006,2008,2009 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/****************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

#include <cutil/avltree.h>

#ifdef  DMALLOC
#include <dmalloc.h>
#endif


/* Internal function prototypes */
int avltree_insert_recurse (AVLTree *tree, AVLNode **nodep, void *data, int *balanced);
void avl_rotate_left(AVLNode **nodep);
void avl_rotate_right(AVLNode **nodep);


/* Basic API functions */
CUTIL_API int avltree_new (AVLTree **treep,
		AVLNodeCompare compare,
		AVLNodeDestroy destroy)
{
	AVLTree *t = NULL;
	if (treep == NULL)
		return EINVAL;
	*treep = NULL;
	t = calloc (1, sizeof (AVLTree));
	if (t == NULL)
		return ENOMEM;
	t->compare = compare;
	t->destroy = destroy;
	*treep = t;
	return ESUCCESS;
}


CUTIL_API int avltree_free (AVLTree ** treep)
{
	AVLTree *t = NULL;
	if (treep == NULL || *treep == NULL)
		return EINVAL;
	t = *treep;
	avltree_remove_left (t, NULL);
	free (t);
	*treep = NULL;
	return ESUCCESS;
}


int avltree_insert_recurse (AVLTree *tree, AVLNode **nodep, void *data, int *balanced)
{
	int	rc;
	int	cmpval;
	AVLNode	*node = NULL;
	AVLNode	*cn = NULL;
	if (tree == NULL || data == NULL || balanced == NULL)
		return EINVAL;
	if (*nodep == NULL)
		return EINVAL;
	cn = *nodep;
	cmpval = tree->compare (data, cn->data);
	if (cmpval < 0) {
		if (avlnode_left(cn) != NULL) {
			rc = avltree_insert_recurse (tree,
						&avlnode_left(cn),
						data,
						balanced);
			if (rc != ESUCCESS)
				return rc;
		} else {
			node = calloc (1, sizeof (AVLNode));
			if (node == NULL) {
				return ENOMEM;
			}
			node->data = data;
			node->factor = AVL_BALANCED;
			avlnode_left(cn) = node;
			*balanced = 0;
		}
		if (!(*balanced)) {
			switch ((*nodep)->factor) {
				case AVL_LEFT_HEAVY:
					avl_rotate_left (nodep);
					*balanced = 1;
					break;
				case AVL_RIGHT_HEAVY:
					cn->factor = AVL_BALANCED;
					*balanced = 1;
					break;
				case AVL_BALANCED:
					cn->factor = AVL_LEFT_HEAVY;
					break;
			}
		}
	} else if (cmpval > 0) {
		if (avlnode_right(cn) != NULL) {
			rc = avltree_insert_recurse (tree,
						&avlnode_right(cn),
						data,
						balanced);
			if (rc != ESUCCESS)
				return rc;
		} else {
			node = calloc (1, sizeof (AVLNode));
			if (node == NULL) {
				return ENOMEM;
			}
			node->data = data;
			node->factor = AVL_BALANCED;
			avlnode_right(cn) = node;
			*balanced = 0;
		}
		if (!(*balanced)) {
			switch (cn->factor) {
				case AVL_RIGHT_HEAVY:
					avl_rotate_right (nodep);
					*balanced = 1;
					break;
				case AVL_LEFT_HEAVY:
					cn->factor = AVL_BALANCED;
					*balanced = 1;
					break;
				case AVL_BALANCED:
					cn->factor = AVL_RIGHT_HEAVY;
					break;
			}
		}
	} else {
		if (cn->hidden != AVL_NODE_HIDDEN)
			return EEXIST;
		cn->hidden = 0;
		cn->data = data;
	}
	return ESUCCESS;
}


CUTIL_API int avltree_insert (AVLTree *tree, void *data)
{
	int	rc;
	int	balanced;
	AVLNode	*node = NULL;
	if (tree == NULL || data == NULL)
		return EINVAL;
	if (tree->root == NULL) {
		node = calloc (1, sizeof (AVLNode));
		if (node == NULL) {
			return ENOMEM;
		}
		node->data = data;
		node->factor = AVL_BALANCED;
		tree->root = node;
		rc = ESUCCESS;
	} else {
		balanced = 0;
		rc = avltree_insert_recurse (tree, &tree->root, data, &balanced);
	}
	return rc;
}


CUTIL_API int avltree_remove (AVLTree *tree, const void *data)
{
	int	cmpval;
	AVLNode	*node = NULL;
	if (tree == NULL || data == NULL)
		return EINVAL;
	node = tree->root;
	while (node != NULL) {
		cmpval = tree->compare (data, node->data);
		if (cmpval < 0) {
			node = node->left;
		} else if (cmpval > 0) {
			node = node->right;
		} else {
			node->hidden = AVL_NODE_HIDDEN;
			return ESUCCESS;
		}
	}

	return ENOENT;
}


CUTIL_API void *avltree_find (AVLTree *tree, void *data)
{
	int	cmpval;
	AVLNode	*node = NULL;
	if (tree == NULL || data == NULL)
		return NULL;
	node = tree->root;
	while (node != NULL) {
		cmpval = tree->compare (data, node->data);
		if (cmpval < 0) {
			node = node->left;
		} else if (cmpval > 0) {
			node = node->right;
		} else {
			if (node->hidden == AVL_NODE_HIDDEN)
				return NULL;
			else
				return node->data;
		}
	}
	return NULL;
}


CUTIL_API int avltree_exists (AVLTree *tree, void *data)
{
	if (avltree_find (tree, data) != NULL)
		return 1;
	return 0;
}


CUTIL_API void *avltree_find_by_compare (AVLTree *tree,
				AVLNodeCompare compare,
				void *data)
{
	int	cmpval;
	AVLNode	*node = NULL;
	if (tree == NULL || compare == NULL || data == NULL)
		return NULL;
	node = tree->root;
	while (node != NULL) {
		cmpval = compare (data, node->data);
		if (cmpval < 0) {
			node = node->left;
		} else if (cmpval > 0) {
			node = node->right;
		} else {
			if (node->hidden == AVL_NODE_HIDDEN)
				return NULL;
			else
				return node->data;
		}
	}
	return NULL;
}


CUTIL_API int avltree_exists_by_compare (AVLTree *tree,
				AVLNodeCompare compare,
				void *data)
{
	if (avltree_find_by_compare (tree, compare, data) != NULL)
		return 1;
	return 0;
}


/* Extended API functions */
CUTIL_API int avltree_insert_left (AVLTree *tree, AVLNode *node, void *data)
{
	AVLNode *n = NULL;
	AVLNode **npos = NULL;

	if (tree == NULL)
		return EINVAL;
	if (node == NULL) {
		if (tree->root != NULL)
			return EINVAL;
	} else {
		npos = &tree->root;
	}
	if (node->left != NULL)
		return EINVAL;
	npos = &node->left;
	n = calloc (1, sizeof (AVLNode));
	if (n == NULL)
		return ENOMEM;
	n->data = data;
	n->factor = AVL_BALANCED;
	*npos = n;
	tree->size++;
	return ESUCCESS;
}


CUTIL_API int avltree_insert_right (AVLTree *tree, AVLNode *node, void *data)
{
	AVLNode *n = NULL;
	AVLNode **npos = NULL;

	if (tree == NULL)
		return EINVAL;
	if (node == NULL) {
		if (tree->root != NULL)
			return EINVAL;
	} else {
		npos = &tree->root;
	}
	if (node->right != NULL)
		return EINVAL;
	npos = &node->right;
	n = calloc (1, sizeof (AVLNode));
	if (n == NULL)
		return ENOMEM;
	n->data = data;
	n->factor = AVL_BALANCED;
	*npos = n;
	tree->size++;
	return ESUCCESS;
}


CUTIL_API int avltree_remove_left (AVLTree *tree, AVLNode *node)
{
	AVLNode **npos = NULL;

	if (tree == NULL)
		return EINVAL;
	if (node == NULL)
		npos = &tree->root;
	else
		npos = &node->left;
	if (*npos != NULL) {
		avltree_remove_left (tree, *npos);
		avltree_remove_right (tree, *npos);
		if (tree->destroy != NULL)
			tree->destroy ((*npos)->data);
		free (*npos);
		*npos = NULL;
		tree->size--;
	}
	return ESUCCESS;
}


CUTIL_API int avltree_remove_right (AVLTree *tree, AVLNode *node)
{
	AVLNode **npos = NULL;

	if (tree == NULL)
		return EINVAL;
	if (node == NULL)
		npos = &tree->root;
	else
		npos = &node->right;
	if (*npos != NULL) {
		avltree_remove_left (tree, *npos);
		avltree_remove_right (tree, *npos);
		if (tree->destroy != NULL)
			tree->destroy ((*npos)->data);
		free (*npos);
		*npos = NULL;
		tree->size--;
	}
	return ESUCCESS;
}



/*
int avltree_size (AVLTree *tree)
{
	if (tree == NULL)
		return 0;
	return tree->size
}


AVLNode *avltree_root (AVLTree *tree)
{
	if (tree == NULL)
		return NULL;
	return tree->root;
}


int avlnode_is_leaf (AVLNode *node)
{
	if (node == NULL)
		return EINVAL;
	return ESUCCESS;
}


AVLNode *avlnode_left (AVLNode *node)
{
	if (node == NULL)
		return NULL;
	return node->left;
}


AVLNode *avlnode_right (AVLNode *node)
{
	if (node == NULL)
		return NULL;
	return node->right;
}


void *avlnode_data (AVLNode *node)
{
	if (node == NULL)
		return NULL;
	return node->data;
}
*/


/* Internal function definitions */
void avl_rotate_left(AVLNode **nodep)
{
	AVLNode	*node = NULL;
	AVLNode	*left = NULL;
	AVLNode *grandchild = NULL;
	if (nodep == NULL || *nodep == NULL)
		return;
	node = *nodep;
	left = avlnode_left (node);
	if (left->factor == AVL_LEFT_HEAVY) {
		avlnode_left(node) = avlnode_right(left);
		avlnode_right(left) = node;
		node->factor = AVL_BALANCED;
		left->factor = AVL_BALANCED;
		*nodep = left;
	} else {
		grandchild = avlnode_right(left);
		avlnode_right(left) = avlnode_left(grandchild);
		avlnode_left(grandchild) = left;
		avlnode_left(node) = avlnode_right(grandchild);
		avlnode_right(grandchild) = node;
		switch (grandchild->factor) {
			case AVL_LEFT_HEAVY:
				node->factor = AVL_RIGHT_HEAVY;
				left->factor = AVL_BALANCED;
				break;
			case AVL_BALANCED:
				node->factor = AVL_BALANCED;
				left->factor = AVL_BALANCED;
				break;
			case AVL_RIGHT_HEAVY:
				node->factor = AVL_BALANCED;
				left->factor = AVL_LEFT_HEAVY;
				break;
		}
		grandchild->factor = AVL_BALANCED;
		*nodep = grandchild;
	}
	return;
}


void avl_rotate_right(AVLNode **nodep)
{
	AVLNode	*node = NULL;
	AVLNode	*right = NULL;
	AVLNode *grandchild = NULL;
	if (nodep == NULL || *nodep == NULL)
		return;
	node = *nodep;
	right = avlnode_right (node);
	if (right->factor == AVL_RIGHT_HEAVY) {
		avlnode_right(node) = avlnode_left(right);
		avlnode_left(right) = node;
		node->factor = AVL_BALANCED;
		right->factor = AVL_BALANCED;
		*nodep = right;
	} else {
		grandchild = avlnode_left(right);
		avlnode_left(right) = avlnode_right(grandchild);
		avlnode_right(grandchild) = right;
		avlnode_right(node) = avlnode_left(grandchild);
		avlnode_left(grandchild) = node;
		switch (grandchild->factor) {
			case AVL_LEFT_HEAVY:
				node->factor = AVL_BALANCED;
				right->factor = AVL_RIGHT_HEAVY;
				break;
			case AVL_BALANCED:
				node->factor = AVL_BALANCED;
				right->factor = AVL_BALANCED;
				break;
			case AVL_RIGHT_HEAVY:
				node->factor = AVL_LEFT_HEAVY;
				right->factor = AVL_BALANCED;
				break;
		}
		grandchild->factor = AVL_BALANCED;
		*nodep = grandchild;
	}
	return;
}


